from torch import nn
import torch
import math
from torch.nn.functional import dropout, linear, pad
from torch.overrides import (
    has_torch_function, has_torch_function_unary, has_torch_function_variadic,
    handle_torch_function)
from torch.nn.init import xavier_uniform_, constant_
from torch.nn import Linear, Dropout, BatchNorm1d

import warnings

from typing import Dict, Optional, Tuple, Union, Callable
import torch.nn.functional as F
from torch import Tensor, nn
from torch.nn import Parameter
from torch.nn.init import xavier_normal_


class MultiheadAttention(nn.Module):

    __constants__ = ['batch_first']
    bias_k: Optional[torch.Tensor]
    bias_v: Optional[torch.Tensor]

    def __init__(self, embed_dim, num_heads, dropout=0., bias=True, add_bias_kv=False, add_zero_attn=False,
                 kdim=None, vdim=None, batch_first=False, device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super(MultiheadAttention, self).__init__()
        self.head_dim = embed_dim // num_heads
        self.embed_dim = embed_dim
        self.kdim = kdim if kdim is not None else embed_dim
        self.vdim = vdim if vdim is not None else embed_dim
        self._qkv_same_embed_dim = self.kdim == embed_dim and self.vdim == embed_dim

        self.num_heads = num_heads
        self.dropout = dropout
        self.batch_first = batch_first
        self.head_dim = embed_dim // num_heads
        assert self.head_dim * num_heads == self.embed_dim, "embed_dim must be divisible by num_heads"
        
        if not self._qkv_same_embed_dim:
            self.q_proj_weight = Parameter(torch.empty((embed_dim, embed_dim), **factory_kwargs))
            self.k_proj_weight = Parameter(torch.empty((embed_dim, self.kdim), **factory_kwargs))
            self.v_proj_weight = Parameter(torch.empty((embed_dim, self.vdim), **factory_kwargs))
            self.register_parameter('in_proj_weight', None)
        else:
            self.in_proj_weight = Parameter(torch.empty((3 * embed_dim, embed_dim), **factory_kwargs))
            self.register_parameter('q_proj_weight', None)
            self.register_parameter('k_proj_weight', None)
            self.register_parameter('v_proj_weight', None)

        if bias:
            self.in_proj_bias = Parameter(torch.empty(3 * embed_dim, **factory_kwargs))
        else:
            self.register_parameter('in_proj_bias', None)
        self.out_proj = nn.Linear(embed_dim, embed_dim, bias=bias)

        if add_bias_kv:
            self.bias_k = Parameter(torch.empty((1, 1, embed_dim), **factory_kwargs))
            self.bias_v = Parameter(torch.empty((1, 1, embed_dim), **factory_kwargs))
        else:
            self.bias_k = self.bias_v = None

        self.add_zero_attn = add_zero_attn
        
        self._reset_parameters()
        self.attn = nn.ModuleList([self.FMM_transfomrer_Attention(attention_dropout=self.dropout, head_dim=self.head_dim,
                                                      diag_size=5, num_head=1, kernels=["elu", "elu_flip"], sparse_ratio=4.5) for i in range(num_heads)])
        
    def _reset_parameters(self):
        if self._qkv_same_embed_dim:
            xavier_uniform_(self.in_proj_weight)
        else:
            xavier_uniform_(self.q_proj_weight)
            xavier_uniform_(self.k_proj_weight)
            xavier_uniform_(self.v_proj_weight)

        if self.in_proj_bias is not None:
            constant_(self.in_proj_bias, 0.)
            constant_(self.out_proj.bias, 0.)
        if self.bias_k is not None:
            xavier_normal_(self.bias_k)
        if self.bias_v is not None:
            xavier_normal_(self.bias_v)

    def __setstate__(self, state):
        # Support loading old MultiheadAttention checkpoints generated by v1.1.0
        if '_qkv_same_embed_dim' not in state:
            state['_qkv_same_embed_dim'] = True

        super(MultiheadAttention, self).__setstate__(state)

    def forward(self, query: Tensor, key: Tensor, value: Tensor, key_padding_mask: Optional[Tensor] = None,
                need_weights: bool = False, attn_mask: Optional[Tensor] = None,
                average_attn_weights: bool = False) -> Tuple[Tensor, Optional[Tensor]]:

        is_batched = query.dim() == 3
        why_not_fast_path = ''
        if not is_batched:
            why_not_fast_path = f"input not batched; expected query.dim() of 3 but got {query.dim()}"
        elif query is not key or key is not value:
            # When lifting this restriction, don't forget to either
            # enforce that the dtypes all match or test cases where
            # they don't!
            why_not_fast_path = "non-self attention was used (query, key, and value are not the same Tensor)"
        elif self.in_proj_bias is not None and query.dtype != self.in_proj_bias.dtype:
            why_not_fast_path = f"dtypes of query ({query.dtype}) and self.in_proj_bias ({self.in_proj_bias.dtype}) don't match"
        elif self.in_proj_weight is not None and query.dtype != self.in_proj_weight.dtype:
            # this case will fail anyway, but at least they'll get a useful error message.
            why_not_fast_path = f"dtypes of query ({query.dtype}) and self.in_proj_weight ({self.in_proj_weight.dtype}) don't match"
        elif self.training:
            why_not_fast_path = "training is enabled"
        elif not self.batch_first:
            why_not_fast_path = "batch_first was not True"
        elif self.bias_k is not None:
            why_not_fast_path = "self.bias_k was not None"
        elif self.bias_v is not None:
            why_not_fast_path = "self.bias_v was not None"
        elif self.dropout:
            why_not_fast_path = f"dropout was {self.dropout}, required zero"
        elif self.add_zero_attn:
            why_not_fast_path = "add_zero_attn was enabled"
        elif not self._qkv_same_embed_dim:
            why_not_fast_path = "_qkv_same_embed_dim was not True"
        elif attn_mask is not None:
            why_not_fast_path = "attn_mask was not None"

        if not why_not_fast_path:
            tensor_args = (
                query,
                key,
                value,
                self.in_proj_weight,
                self.in_proj_bias,
                self.out_proj.weight,
                self.out_proj.bias,
            )
            # We have to use list comprehensions below because TorchScript does not support
            # generator expressions.
            if torch.overrides.has_torch_function(tensor_args):
                why_not_fast_path = "some Tensor argument has_torch_function"
            elif not all([(x.is_cuda or 'cpu' in str(x.device)) for x in tensor_args]):
                why_not_fast_path = "some Tensor argument is neither CUDA nor CPU"
            elif torch.is_grad_enabled() and any([x.requires_grad for x in tensor_args]):
                why_not_fast_path = ("grad is enabled and at least one of query or the "
                                     "input/output projection weights or biases requires_grad")
            if not why_not_fast_path:
                assert 1==2, 'what is why_not_fast_path'
                return torch._native_multi_head_attention(
                    query,
                    key,
                    value,
                    self.embed_dim,
                    self.num_heads,
                    self.in_proj_weight,
                    self.in_proj_bias,
                    self.out_proj.weight,
                    self.out_proj.bias,
                    key_padding_mask if key_padding_mask is not None else attn_mask,
                    need_weights,
                    average_attn_weights)

        if self.batch_first and is_batched:
            # make sure that the transpose op does not affect the "is" property
            if key is value:
                if query is key:
                    query = key = value = query.transpose(1, 0)
                else:
                    query, key = [x.transpose(1, 0) for x in (query, key)]
                    value = key
            else:
                query, key, value = [x.transpose(1, 0) for x in (query, key, value)]

        if not self._qkv_same_embed_dim:
            attn_output, attn_output_weights = self.Mrs_multi_head_attention_forward(
                query, key, value, self.embed_dim, self.num_heads,
                self.in_proj_weight, self.in_proj_bias,
                self.bias_k, self.bias_v, self.add_zero_attn,
                self.dropout, self.out_proj.weight, self.out_proj.bias,
                training=self.training,
                key_padding_mask=key_padding_mask, need_weights=need_weights,
                attn_mask=attn_mask, use_separate_proj_weight=True,
                q_proj_weight=self.q_proj_weight, k_proj_weight=self.k_proj_weight,
                v_proj_weight=self.v_proj_weight, average_attn_weights=average_attn_weights)
        else:
            attn_output, attn_output_weights = self.Mrs_multi_head_attention_forward(
                query, key, value, self.embed_dim, self.num_heads,
                self.in_proj_weight, self.in_proj_bias,
                self.bias_k, self.bias_v, self.add_zero_attn,
                self.dropout, self.out_proj.weight, self.out_proj.bias,
                training=self.training,
                key_padding_mask=key_padding_mask, need_weights=need_weights,
                attn_mask=attn_mask, average_attn_weights=average_attn_weights,
                head_dim=self.head_dim)
            
        if self.batch_first and is_batched:
            return attn_output.transpose(1, 0), attn_output_weights
        else:
            return attn_output, attn_output_weights

    def Mrs_multi_head_attention_forward(self,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        embed_dim_to_check: int,
        num_heads: int,
        in_proj_weight: Optional[Tensor],
        in_proj_bias: Optional[Tensor],
        bias_k: Optional[Tensor],
        bias_v: Optional[Tensor],
        add_zero_attn: bool,
        dropout_p: float,
        out_proj_weight: Tensor,
        out_proj_bias: Optional[Tensor],
        training: bool = True,
        key_padding_mask: Optional[Tensor] = None,
        need_weights: bool = True,
        attn_mask: Optional[Tensor] = None,
        use_separate_proj_weight: bool = False,
        q_proj_weight: Optional[Tensor] = None,
        k_proj_weight: Optional[Tensor] = None,
        v_proj_weight: Optional[Tensor] = None,
        static_k: Optional[Tensor] = None,
        static_v: Optional[Tensor] = None,
        average_attn_weights: bool = True,
        head_dim=None,
    ):    
        is_batched = _mha_shape_check(query, key, value, key_padding_mask, attn_mask, num_heads)
    
        # For unbatched input, we unsqueeze at the expected batch-dim to pretend that the input
        # is batched, run the computation and before returning squeeze the
        # batch dimension so that the output doesn't carry this temporary batch dimension.
        if not is_batched:
            # unsqueeze if the input is unbatched
            query = query.unsqueeze(1)
            key = key.unsqueeze(1)
            value = value.unsqueeze(1)
            if key_padding_mask is not None:
                key_padding_mask = key_padding_mask.unsqueeze(0)

        tgt_len, bsz, embed_dim = query.shape
        src_len, _, _ = key.shape
    
        #
        # compute in-projection
        #
        # query: [seq_len, bsz, dim]
        group_by_list = [1, 1, 2, 2, 4, 4, 8, 8]
        X = query
        seq_len, batch_size, dim = X.shape
        if attn_mask is not None:
            X = X * attn_mask[:, :, None]
        attn_output = torch.empty(batch_size, num_heads, seq_len, dim // num_heads, device=X.device)
        for h in range(num_heads):
            X_ = F.avg_pool1d(X.transpose(0, -1), kernel_size=group_by_list[h], stride=group_by_list[h], ceil_mode=True, count_include_pad=False).transpose(0, -1)
            seq_len_, batch_size, dim = X_.shape
            # Down sampling mask and input
            if attn_mask is not None:
                attn_mask_ = torch.clip(attn_mask.reshape(X.shape[0], X.shape[1] // group_by_list[h], group_by_list[h]).sum(dim=-1), min=0, max=1)
            if not use_separate_proj_weight:
                assert in_proj_weight is not None, "use_separate_proj_weight is False but in_proj_weight is None"
                q_, k_, v_ = _in_projection_packed(X_, X_, X_, in_proj_weight, in_proj_bias)
            else:
                assert q_proj_weight is not None, "use_separate_proj_weight is True but q_proj_weight is None"
                assert k_proj_weight is not None, "use_separate_proj_weight is True but k_proj_weight is None"
                assert v_proj_weight is not None, "use_separate_proj_weight is True but v_proj_weight is None"
                if in_proj_bias is None:
                    b_q = b_k = b_v = None
                else:
                    b_q, b_k, b_v = in_proj_bias.chunk(3)
                q_, k_, v_ = _in_projection(query, key, value, q_proj_weight, k_proj_weight, v_proj_weight, b_q, b_k, b_v)
        
            # prep key padding mask
            if key_padding_mask is not None and key_padding_mask.dtype == torch.uint8:
                warnings.warn("Byte tensor for key_padding_mask in nn.MultiheadAttention is deprecated. Use bool tensor instead.")
                key_padding_mask = key_padding_mask.to(torch.bool)
        
            # add bias along batch dimension (currently second)
            if bias_k is not None and bias_v is not None:
                assert static_k is None, "bias cannot be added to static key."
                assert static_v is None, "bias cannot be added to static value."
                k_ = torch.cat([k_, bias_k.repeat(1, bsz, 1)])
                v_ = torch.cat([v_, bias_v.repeat(1, bsz, 1)])
                if key_padding_mask is not None:
                    key_padding_mask = pad(key_padding_mask, (0, 1))
            else:
                assert bias_k is None
                assert bias_v is None
        
            # adjust dropout probability
            if not training:
                dropout_p = 0.0
        
            # (deep breath) calculate attention and out projection
            # [bsz, seq_len, hidden_dim] -> [bsz, seq_len, num_head, head_dim]
            q_ = q_.contiguous().view(seq_len_, batch_size, num_heads, head_dim).transpose(0, 1)[:, :, h, :].unsqueeze(-2)
            k_ = k_.contiguous().view(seq_len_, batch_size, num_heads, head_dim).transpose(0, 1)[:, :, h, :].unsqueeze(-2)
            v_ = v_.contiguous().view(seq_len_, batch_size, num_heads, head_dim).transpose(0, 1)[:, :, h, :].unsqueeze(-2)
            attn_output_, attn_output_weights = self.attn[h](q_, k_, v_, attn_mask, dropout_p)
            attn_output[:, h, :, :] = attn_output_.squeeze().repeat_interleave(group_by_list[h], dim=1)[:, :seq_len, :]
        # bsz, n_head, seq_len, hidden_dim
        attn_output = attn_output.permute(2, 0, 1, 3).contiguous().view(tgt_len * bsz, embed_dim)
        attn_output = linear(attn_output, out_proj_weight, out_proj_bias)
        attn_output = attn_output.view(tgt_len, bsz, attn_output.size(1))
    
    
        if need_weights:
            # optionally average attention weights over heads
            if average_attn_weights:
                attn_output_weights = attn_output_weights.sum(dim=1) / num_heads
    
            if not is_batched:
                # squeeze the output if input was unbatched
                attn_output = attn_output.squeeze(1)
                attn_output_weights = attn_output_weights.squeeze(0)
            return attn_output, attn_output_weights
        else:
            if not is_batched:
                # squeeze the output if input was unbatched
                attn_output = attn_output.squeeze(1)
    #         print(attn_output)
            return attn_output, None

    class FMM_transfomrer_Attention(nn.Module):
        def __init__(self, attention_dropout, head_dim, diag_size, num_head, kernels, sparse_ratio):
            super(MultiheadAttention.FMM_transfomrer_Attention, self).__init__()
            self.dropout = nn.Dropout(attention_dropout)
            self.head_dim = head_dim
            self.diag_size = diag_size
            self.num_head = num_head
            self.sparse_ratio = sparse_ratio
            
            self.kernels = kernels
            self.rank_k = len(self.kernels)
            
            if self.sparse_ratio < 1.0: # e.g. sparse_ratio = 0.5
                self.type_blend = 0
            elif self.sparse_ratio < 2.0: # e.g. sparse_ratio = 1.5
                self.type_blend = 1
            elif self.sparse_ratio < 3.0: # e.g. sparse_ratio = 2.5
                self.type_blend = 2
            elif self.sparse_ratio < 4.0:  # e.g. sparse_ratio = 3.5
                self.type_blend = 3
            elif self.sparse_ratio < 5.0: # e.g. sparse_ratio = 4.5
                self.type_blend = 4 
            elif self.sparse_ratio < 6.0: # e.g. sparse_ratio = 5.5
                self.type_blend = 5 
            elif self.sparse_ratio < 7.0: # e.g. sparse_ratio = 6.5
                self.type_blend = 6 # sparse only
            elif self.sparse_ratio < 8.0: # e.g. sparse_ratio = 7.5
                self.type_blend = 7 # lowrank only 
            elif self.sparse_ratio < 9.0: # e.g. sparse_ratio = 8.5
                self.type_blend = 8 
            elif self.sparse_ratio < 10.0: # e.g. sparse_ratio = 9.5
                self.type_blend = 9 
            elif self.sparse_ratio < 11.0: # e.g. sparse_ratio = 10.5
                self.type_blend = 10 
            elif self.sparse_ratio < 12.0: # e.g. sparse_ratio = 11.5
                self.type_blend = 11 
            elif self.sparse_ratio < 13.0: # e.g. sparse_ratio = 12.5
                self.type_blend = 12 
            elif self.sparse_ratio < 14.0: # e.g. sparse_ratio = 13.5
                self.type_blend = 13
            elif self.sparse_ratio < 15.0: # e.g. sparse_ratio = 14.5
                self.type_blend = 14 
            elif self.sparse_ratio < 16.0: # e.g. sparse_ratio = 15.5
                self.type_blend = 15 
            elif self.sparse_ratio < 17.0: # e.g. sparse_ratio = 16.5
                self.type_blend = 16
            elif self.sparse_ratio < 18.0: # e.g. sparse_ratio = 17.5
                self.type_blend = 17
            elif self.sparse_ratio < 19.0: # e.g. sparse_ratio = 18.5
                self.type_blend = 18
            elif self.sparse_ratio < 20.0: # e.g. sparse_ratio = 19.5
                self.type_blend = 19
                
            if self.type_blend == 0:
                self.sparse_ratio = nn.Parameter(torch.Tensor([0.5]))
            elif self.type_blend == 1:
                self.sparse_ratio = nn.Parameter(torch.Tensor([0.5]))
            elif self.type_blend == 2:
                self.sparse_ratio = nn.Parameter(torch.Tensor([0.5]))
            elif self.type_blend == 3:
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, self.head_dim))
            elif self.type_blend == 4:  
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, self.head_dim))
            elif self.type_blend == 5:  
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, self.head_dim))
            elif self.type_blend == 6:
                self.sparse_ratio = 1.0 # sparse only
            elif self.type_blend == 7:
                self.sparse_ratio = 0.0 # lowrank only
            elif self.type_blend == 8:
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, self.head_dim))
                self.lowrank_ratio = nn.Parameter(torch.Tensor([0.5]))
                self.lowrank_ratio2 = nn.Parameter(torch.Tensor([0.5]))
            elif self.type_blend == 9:
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, self.head_dim))
                self.lowrank_ratio = nn.Parameter(torch.Tensor([0.5]))
                self.lowrank_ratio2 = nn.Parameter(torch.Tensor([0.5]))
            elif self.type_blend == 10:
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, 1))
                self.sparse_ratio2 = nn.Parameter(torch.ones(1, self.num_head, 1, 1))
            elif self.type_blend == 11:
                self.sparse_ratio = nn.Parameter(torch.ones(1, self.num_head, 1, 1))
                self.sparse_ratio2 = nn.Parameter(torch.zeros(1, self.num_head, 1, 1))
            elif self.type_blend == 12:
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, 1))
            elif self.type_blend == 13:
                self.sparse_ratio = nn.Parameter(torch.ones(1, self.num_head, 1, 1))
                self.sparse_ratio2 = nn.Parameter(torch.zeros(1, self.num_head, 1, 1))
                self.lowrank_ratio = nn.Parameter(torch.Tensor([1.0]))
                self.lowrank_ratio2 = nn.Parameter(torch.Tensor([0.0]))
            elif self.type_blend == 14:
                self.sparse_ratio = nn.Parameter(torch.Tensor([0.5]))
                self.lowrank_ratio = nn.Parameter(torch.Tensor([1.0]))
                self.lowrank_ratio2 = nn.Parameter(torch.Tensor([0.0]))
            elif self.type_blend == 15:
                self.sparse_ratio = nn.Parameter(torch.Tensor([0.5]))
                self.lowrank_ratio = nn.Parameter(torch.Tensor([0.0]))
            elif self.type_blend == 16:  
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, self.head_dim))
                self.lowrank_ratio = nn.Parameter(torch.Tensor([1.0]))
                self.lowrank_ratio2 = nn.Parameter(torch.Tensor([0.0]))
            elif self.type_blend == 17:  
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, self.head_dim))
                self.lowrank_ratio = nn.Parameter(torch.Tensor([0.0]))
            elif self.type_blend == 18:
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, 1))
            elif self.type_blend == 19:
                self.sparse_ratio = nn.Parameter(torch.zeros(1, self.num_head, 1, 1))
                self.lowrank_ratio = nn.Parameter(torch.Tensor([1.0]))
                self.lowrank_ratio2 = nn.Parameter(torch.Tensor([0.0]))
            
        def forward(self, Q, K, V, mask, dropout_p=0.0):
            # [bsz, seq_len, no_heads, hidden_dim]
            Q = Q.transpose(1, 2)
            K = K.transpose(1, 2)
            V = V.transpose(1, 2)
            if self.type_blend != 7:
                attn_vec_sparse = self._forward_sparse(Q, K, V, mask) # batchsize x num_head x seq.len. x head_dim
        
            if self.type_blend != 6:
                attn_vec_lowrank = self._forward_lowrank(Q, K, V, mask) # batchsize x num_head x seq.len. x head_dim
                
            if self.type_blend == 0:
                X = self.sparse_ratio * attn_vec_sparse + attn_vec_lowrank
            elif self.type_blend == 1:
                X = attn_vec_sparse + self.sparse_ratio * attn_vec_lowrank
            elif self.type_blend == 2:
                X = self.sparse_ratio * attn_vec_sparse + (1.0 - self.sparse_ratio) * attn_vec_lowrank
            elif self.type_blend == 3:
                X = self.sparse_ratio * attn_vec_sparse + attn_vec_lowrank
            elif self.type_blend == 4:
                X = attn_vec_sparse + self.sparse_ratio * attn_vec_lowrank
            elif self.type_blend == 5:
                X = self.sparse_ratio * attn_vec_sparse + (1.0 - self.sparse_ratio) * attn_vec_lowrank
            elif self.type_blend == 6:
                X = attn_vec_sparse
            elif self.type_blend == 7:
                X = attn_vec_lowrank
            elif self.type_blend == 8:
                X = self.sparse_ratio * attn_vec_sparse + attn_vec_lowrank
            elif self.type_blend == 9:
                X = attn_vec_sparse + self.sparse_ratio * attn_vec_lowrank
            elif self.type_blend == 10:
                X = self.sparse_ratio * attn_vec_sparse + self.sparse_ratio2 * attn_vec_lowrank
            elif self.type_blend == 11:
                X = self.sparse_ratio * attn_vec_sparse + self.sparse_ratio2 * attn_vec_lowrank
            elif self.type_blend == 12:
                X = attn_vec_sparse + self.sparse_ratio * attn_vec_lowrank
            elif self.type_blend == 13:
                X = self.sparse_ratio * attn_vec_sparse + self.sparse_ratio2 * attn_vec_lowrank
            elif self.type_blend == 14:
                X = self.sparse_ratio * attn_vec_sparse + (1.0 - self.sparse_ratio) * attn_vec_lowrank
            elif self.type_blend == 15:
                X = self.sparse_ratio * attn_vec_sparse + (1.0 - self.sparse_ratio) * attn_vec_lowrank
            elif self.type_blend == 16:
                X = attn_vec_sparse + self.sparse_ratio * attn_vec_lowrank
            elif self.type_blend == 17:
                X = attn_vec_sparse + self.sparse_ratio * attn_vec_lowrank
            elif self.type_blend == 18:
                X = self.sparse_ratio * attn_vec_sparse + attn_vec_lowrank
            elif self.type_blend == 19:
                X = self.sparse_ratio * attn_vec_sparse + attn_vec_lowrank
            
            return X, None
        
        def _project_features(self, features, kernel_name):
            if kernel_name == 'elu':
                out = F.elu(features, 1., False) + 1.
            elif kernel_name == 'tanh':
                out = F.tanh(features) + 1.
            elif kernel_name == 'relu':
                out = F.relu(features, False)
            elif kernel_name == 'celu':
                out = F.celu(features, 1., False) + 1.
            elif kernel_name == 'sigmoid':
                out = F.sigmoid(features)
            elif kernel_name == 'leaky_relu':
                out = F.leaky_relu(features) + 1.
            elif kernel_name == 'softplus':
                out = F.softplus(features)
            elif kernel_name == 'tanh_orthogonal':
                out = 1. - F.tanh(features)
            elif kernel_name == 'elu_flip':
                out = F.elu(-features, 1., False) + 1.
            else:
                out = features
                
            return out
                
        def _forward_lowrank(self, Q, K, V, mask):
            if mask is not None:
                V = V * mask[:, None, :, None]
            
            Q1 = self._project_features(Q, kernel_name=self.kernels[0]) / math.sqrt(math.sqrt(Q.size(2)))
            if mask is not None:
                K1 = self._project_features(K, kernel_name=self.kernels[0]) * mask[:, None, :, None] / math.sqrt(math.sqrt(K.size(2)))
            else:
                K1 = self._project_features(K, kernel_name=self.kernels[0]) / math.sqrt(math.sqrt(K.size(2)))
            X = torch.matmul(Q1, torch.matmul(torch.transpose(K1, -2, -1), V))
            
            if self.rank_k > 1:
                Q2 = self._project_features(Q, kernel_name=self.kernels[1]) / math.sqrt(math.sqrt(Q.size(2)))
                if mask is not None:
                    K2 = self._project_features(K, kernel_name=self.kernels[1]) * mask[:, None, :, None] / math.sqrt(math.sqrt(K.size(2)))
                else:
                    K2 = self._project_features(K, kernel_name=self.kernels[1]) / math.sqrt(math.sqrt(K.size(2)))
                X2 = torch.matmul(Q2, torch.matmul(torch.transpose(K2, -2, -1), V))
                
                if self.type_blend == 8:
                    X = self.lowrank_ratio * X + self.lowrank_ratio2 * X2
                elif self.type_blend == 9:
                    X = self.lowrank_ratio * X + self.lowrank_ratio2 * X2
                elif self.type_blend == 13:
                    X = self.lowrank_ratio * X + self.lowrank_ratio2 * X2
                elif self.type_blend == 14:
                    X = self.lowrank_ratio * X + self.lowrank_ratio2 * X2
                elif self.type_blend == 15:
                    X = X + self.lowrank_ratio * X2
                elif self.type_blend == 16:
                    X = self.lowrank_ratio * X + self.lowrank_ratio2 * X2
                elif self.type_blend == 17:
                    X = X + self.lowrank_ratio * X2
                elif self.type_blend == 19:
                    X = self.lowrank_ratio * X + self.lowrank_ratio2 * X2
                else:
                    X = 0.5 * X + 0.5 * X2

            return X
        
        def _forward_sparse(self, Q, K, V, mask):
            dot = torch.matmul(Q, torch.transpose(K, -2, -1))
            dot = dot / math.sqrt(self.head_dim)
            if mask is not None:
                dot = dot - 1e6 * (1 - mask[:, None, None, :])
            
            #### Computing masked softmax(sparse) ####
            bsz, n_head, qlen, klen = dot.shape
            
            sparse_mask = torch.ones(qlen, klen).to(dot)
            sparse_mask = torch.tril(sparse_mask, diagonal=-self.diag_size) + torch.triu(sparse_mask, diagonal=self.diag_size)
            sparse_mask = sparse_mask.to(torch.bool)
            
            dot.masked_fill_(sparse_mask[None, None, :, :], -float('inf'))
            
            attn = nn.functional.softmax(dot, dim = -1)
            attn = self.dropout(attn)

            X = torch.matmul(attn, V)
            return X
    # if attn_mask is not None:
    #     attn *= (attn_mask == 0.)
    # if dropout_p > 0.0:
    #     attn = F.dropout(attn, p=dropout_p)
    # X = torch.matmul(attn, values)
    # return X, attn

def _in_projection_packed(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    w: Tensor,
    b: Optional[Tensor] = None,
):
    E = q.size(-1)
    if k is v:
        if q is k:
            # self-attention
            return linear(q, w, b).chunk(3, dim=-1)
        else:
            # encoder-decoder attention
            w_q, w_kv = w.split([E, E * 2])
            if b is None:
                b_q = b_kv = None
            else:
                b_q, b_kv = b.split([E, E * 2])
            return (linear(q, w_q, b_q),) + linear(k, w_kv, b_kv).chunk(2, dim=-1)
    else:
        w_q, w_k, w_v = w.chunk(3)
        if b is None:
            b_q = b_k = b_v = None
        else:
            b_q, b_k, b_v = b.chunk(3)
        return linear(q, w_q, b_q), linear(k, w_k, b_k), linear(v, w_v, b_v)

def _mha_shape_check(query: Tensor, key: Tensor, value: Tensor,
                     key_padding_mask: Optional[Tensor], attn_mask: Optional[Tensor], num_heads: int):

    if query.dim() == 3:
        # Batched Inputs
        is_batched = True
        assert key.dim() == 3 and value.dim() == 3, \
            ("For batched (3-D) `query`, expected `key` and `value` to be 3-D"
             f" but found {key.dim()}-D and {value.dim()}-D tensors respectively")
        if key_padding_mask is not None:
            assert key_padding_mask.dim() == 2, \
                ("For batched (3-D) `query`, expected `key_padding_mask` to be `None` or 2-D"
                 f" but found {key_padding_mask.dim()}-D tensor instead")
        if attn_mask is not None:
            assert attn_mask.dim() in (2, 3), \
                ("For batched (3-D) `query`, expected `attn_mask` to be `None`, 2-D or 3-D"
                 f" but found {attn_mask.dim()}-D tensor instead")
    elif query.dim() == 2:
        # Unbatched Inputs
        is_batched = False
        assert key.dim() == 2 and value.dim() == 2, \
            ("For unbatched (2-D) `query`, expected `key` and `value` to be 2-D"
             f" but found {key.dim()}-D and {value.dim()}-D tensors respectively")

        if key_padding_mask is not None:
            assert key_padding_mask.dim() == 1, \
                ("For unbatched (2-D) `query`, expected `key_padding_mask` to be `None` or 1-D"
                 f" but found {key_padding_mask.dim()}-D tensor instead")

        if attn_mask is not None:
            assert attn_mask.dim() in (2, 3), \
                ("For unbatched (2-D) `query`, expected `attn_mask` to be `None`, 2-D or 3-D"
                 f" but found {attn_mask.dim()}-D tensor instead")
            if attn_mask.dim() == 3:
                expected_shape = (num_heads, query.shape[0], key.shape[0])
                assert attn_mask.shape == expected_shape, \
                    (f"Expected `attn_mask` shape to be {expected_shape} but got {attn_mask.shape}")
    else:
        raise AssertionError(
            f"query should be unbatched 2D or batched 3D tensor but received {query.dim()}-D query tensor")

    return is_batched

def _in_projection(
    q: Tensor,
    k: Tensor,
    v: Tensor,
    w_q: Tensor,
    w_k: Tensor,
    w_v: Tensor,
    b_q: Optional[Tensor] = None,
    b_k: Optional[Tensor] = None,
    b_v: Optional[Tensor] = None,
):
    Eq, Ek, Ev = q.size(-1), k.size(-1), v.size(-1)
    assert w_q.shape == (Eq, Eq), f"expecting query weights shape of {(Eq, Eq)}, but got {w_q.shape}"
    assert w_k.shape == (Eq, Ek), f"expecting key weights shape of {(Eq, Ek)}, but got {w_k.shape}"
    assert w_v.shape == (Eq, Ev), f"expecting value weights shape of {(Eq, Ev)}, but got {w_v.shape}"
    assert b_q is None or b_q.shape == (Eq,), f"expecting query bias shape of {(Eq,)}, but got {b_q.shape}"
    assert b_k is None or b_k.shape == (Eq,), f"expecting key bias shape of {(Eq,)}, but got {b_k.shape}"
    assert b_v is None or b_v.shape == (Eq,), f"expecting value bias shape of {(Eq,)}, but got {b_v.shape}"
    return linear(q, w_q, b_q), linear(k, w_k, b_k), linear(v, w_v, b_v)


# main model class
class TransformerEncoderClassiregressor(nn.Module):
    """
    Simplest classifier/regressor. Can be either regressor or classifier because the output does not include
    softmax. Concatenates final layer embeddings and uses 0s to ignore padding embeddings in final output layer.
    """

    def __init__(self, feat_dim, max_len, d_model, n_heads, num_layers, dim_feedforward, num_classes,
                 dropout=0.1, pos_encoding='fixed', activation='gelu', norm='BatchNorm', freeze=False):
        super(TransformerEncoderClassiregressor, self).__init__()

        self.max_len = max_len
        self.d_model = d_model
        self.n_heads = n_heads

        self.project_inp = nn.Linear(feat_dim, d_model)
        self.pos_enc = get_pos_encoder(pos_encoding)(d_model, dropout=dropout*(1.0 - freeze), max_len=max_len)

        encoder_layer = TransformerBatchNormEncoderLayer(d_model, self.n_heads, dim_feedforward, dropout*(1.0 - freeze), activation=activation)

        self.transformer_encoder = nn.TransformerEncoder(encoder_layer, num_layers)

        self.act = _get_activation_fn(activation)

        self.dropout1 = nn.Dropout(dropout)

        self.feat_dim = feat_dim
        self.num_classes = num_classes
        self.output_layer = self.build_output_module(d_model, max_len, num_classes)

    def build_output_module(self, d_model, max_len, num_classes):
        output_layer = nn.Linear(d_model * max_len, num_classes)
        # no softmax (or log softmax), because CrossEntropyLoss does this internally. If probabilities are needed,
        # add F.log_softmax and use NLLoss
        return output_layer

    def forward(self, X, padding_masks):
        """
        Args:
            X: (batch_size, seq_length, feat_dim) torch tensor of masked features (input)
            padding_masks: (batch_size, seq_length) boolean tensor, 1 means keep vector at this position, 0 means padding
        Returns:
            output: (batch_size, num_classes)
        """

        # permute because pytorch convention for transformers is [seq_length, batch_size, feat_dim]. padding_masks [batch_size, feat_dim]
        inp = X.permute(1, 0, 2)
        inp = self.project_inp(inp) * math.sqrt(
            self.d_model)  # [seq_length, batch_size, d_model] project input vectors to d_model dimensional space
        inp = self.pos_enc(inp)  # add positional encoding
        # NOTE: logic for padding masks is reversed to comply with definition in MultiHeadAttention, TransformerEncoderLayer
        output = self.transformer_encoder(inp, src_key_padding_mask=~padding_masks)  # (seq_length, batch_size, d_model)
        output = self.act(output)  # the output transformer encoder/decoder embeddings don't include non-linearity
        output = output.permute(1, 0, 2)  # (batch_size, seq_length, d_model)
        output = self.dropout1(output)

        # Output
        output = output * padding_masks.unsqueeze(-1)  # zero-out padding embeddings
        output = output.reshape(output.shape[0], -1)  # (batch_size, seq_length * d_model)
        output = self.output_layer(output)  # (batch_size, num_classes)

        return output

def get_pos_encoder(pos_encoding):
    if pos_encoding == "learnable":
        return LearnablePositionalEncoding
    raise NotImplementedError("pos_encoding should be 'learnable'/'fixed', not '{}'".format(pos_encoding))


class LearnablePositionalEncoding(nn.Module):

    def __init__(self, d_model, dropout=0.1, max_len=1024):
        super(LearnablePositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)
        # Each position gets its own embedding
        # Since indices are always 0 ... max_len, we don't have to do a look-up
        self.pe = nn.Parameter(torch.empty(max_len, 1, d_model))  # requires_grad automatically set to True
        nn.init.uniform_(self.pe, -0.02, 0.02)

    def forward(self, x):
        r"""Inputs of forward function
        Args:
            x: the sequence fed to the positional encoder model (required).
        Shape:
            x: [sequence length, batch size, embed dim]
            output: [sequence length, batch size, embed dim]
        """

        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)


class TransformerBatchNormEncoderLayer(nn.modules.Module):
    r"""This transformer encoder layer block is made up of self-attn and feedforward network.
    It differs from TransformerEncoderLayer in torch/nn/modules/transformer.py in that it replaces LayerNorm
    with BatchNorm.
    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of intermediate layer, relu or gelu (default=relu).
    """

    def __init__(self, d_model, nhead, dim_feedforward=2048, dropout=0.1, activation="relu", **kwargs):
        super(TransformerBatchNormEncoderLayer, self).__init__()
        self.self_attn = MultiheadAttention(d_model, nhead, dropout=dropout, **kwargs)
        # Implementation of Feedforward model
        self.linear1 = Linear(d_model, dim_feedforward)
        self.dropout = Dropout(dropout)
        self.linear2 = Linear(dim_feedforward, d_model)

        self.norm1 = BatchNorm1d(d_model, eps=1e-5)  # normalizes each feature across batch samples and time steps
        self.norm2 = BatchNorm1d(d_model, eps=1e-5)
        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)

        self.activation = _get_activation_fn(activation)

    def __setstate__(self, state):
        if 'activation' not in state:
            state['activation'] = F.relu
        super(TransformerBatchNormEncoderLayer, self).__setstate__(state)

    def forward(self, src: Tensor, src_mask: Optional[Tensor] = None,
                src_key_padding_mask: Optional[Tensor] = None) -> Tensor:
        r"""Pass the input through the encoder layer.
        Args:
            src: the sequence to the encoder layer (required).
            src_mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).
        Shape:
            see the docs in Transformer class.
        """
        src2 = self.self_attn(src, src, src, attn_mask=src_mask,
                              key_padding_mask=src_key_padding_mask, average_attn_weights=False)[0]
        src = src + self.dropout1(src2)  # (seq_len, batch_size, d_model)
        src = src.permute(1, 2, 0)  # (batch_size, d_model, seq_len)
        # src = src.reshape([src.shape[0], -1])  # (batch_size, seq_length * d_model)
        src = self.norm1(src)
        src = src.permute(2, 0, 1)  # restore (seq_len, batch_size, d_model)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)  # (seq_len, batch_size, d_model)
        src = src.permute(1, 2, 0)  # (batch_size, d_model, seq_len)
        src = self.norm2(src)
        src = src.permute(2, 0, 1)  # restore (seq_len, batch_size, d_model)
        return src


def _get_activation_fn(activation: str) -> Callable[[Tensor], Tensor]:
    if activation == "relu":
        return F.relu
    elif activation == "gelu":
        return F.gelu

    raise RuntimeError("activation should be relu/gelu, not {}".format(activation))
